package org.ballerinalang.ballerina.swagger.convertor.service;

import io.swagger.models.Scheme;
import io.swagger.models.Swagger;
import org.ballerinalang.ballerina.swagger.convertor.Constants;
import org.ballerinalang.ballerina.swagger.convertor.ConverterUtils;
import org.ballerinalang.model.tree.NodeKind;
import org.ballerinalang.model.tree.ServiceNode;
import org.ballerinalang.model.tree.expressions.SimpleVariableReferenceNode;
import org.wso2.ballerinalang.compiler.tree.BLangSimpleVariable;
import org.wso2.ballerinalang.compiler.tree.expressions.BLangExpression;
import org.wso2.ballerinalang.compiler.tree.expressions.BLangNamedArgsExpression;
import org.wso2.ballerinalang.compiler.tree.expressions.BLangRecordLiteral;
import org.wso2.ballerinalang.compiler.tree.expressions.BLangTypeInit;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * Extract Swagger server information from and Ballerina endpoint.
 */
public class SwaggerEndpointMapper {

    /**
     * Convert endpoints bound to {@code service} swagger server information.
     * Currently this method only selects last endpoint bound to {@code service}
     * as the Server for swagger definition. This is due to OAS2 not supporting
     * multiple Server definitions.
     *
     * @param endpoints all endpoints defined in ballerina source
     * @param service   service node with bound endpoints
     * @param swagger   swagger definition to attach extracted information
     * @return swagger definition with Server information
     */
    public Swagger convertBoundEndpointsToSwagger(List<BLangSimpleVariable> endpoints, ServiceNode service,
                                                  Swagger swagger) {
        if (endpoints == null || service == null || service.getAttachedExprs().isEmpty()
                || service.getAttachedExprs().get(0).getKind() != NodeKind.SIMPLE_VARIABLE_REF) {
            return swagger;
        }
        if (swagger == null) {
            return new Swagger();
        }

        SimpleVariableReferenceNode node = (SimpleVariableReferenceNode) service.getAttachedExprs().get(0);
        for (BLangSimpleVariable ep : endpoints) {
            // At the moment only the last bound endpoint will be populated in swagger
            // we need to move to OAS3 models to support multiple server support
            if (node.getVariableName().equals(ep.getName())) {
                extractServer(ep, swagger);
                break;
            }
        }
        return swagger;
    }

    private void extractServer(BLangSimpleVariable ep, Swagger swagger) {
        BLangTypeInit bTypeInit = (BLangTypeInit) ep.getInitialExpression();
        List<BLangExpression> list = bTypeInit.argsExpr;
        List<Scheme> schemes = new ArrayList<>();
        String port = null;
        String host = null;

        if (list.size() == 1) {
            port = ConverterUtils.getStringLiteralValue(list.get(0));
        }

        if (list.size() > 1
                && list.get(1) instanceof BLangNamedArgsExpression
                && ((BLangNamedArgsExpression) list.get(1)).getExpression() instanceof BLangRecordLiteral) {
            BLangRecordLiteral bLangRecordLiteral = (BLangRecordLiteral) ((BLangNamedArgsExpression) list.get(1))
                    .getExpression();
            Map<String, BLangExpression> configs = ConverterUtils.listToMap(bLangRecordLiteral.keyValuePairs);

            host = configs.get(Constants.ATTR_HOST) != null ?
                    ConverterUtils.getStringLiteralValue(configs.get(Constants.ATTR_HOST)) :
                    null;
            Scheme scheme = configs.get(Constants.SECURE_SOCKETS) == null ? Scheme.HTTP : Scheme.HTTPS;
            schemes.add(scheme);
        }

        // Set default values to host and port if values are not defined
        if (host == null) {
            host = Constants.ATTR_DEF_HOST;
        }
        if (port != null) {
            host += ':' + port;
        }

        swagger.setHost(host);
        swagger.setSchemes(schemes);
    }
}
